@@ -1,18 +1,22 @@
 """Provide the Subreddit class."""
+# pylint: disable=too-many-lines
 from copy import deepcopy
-from json import dumps
-import time
+from json import dumps, loads
+from os.path import basename, dirname, join
 
 from prawcore import Redirect
+import websocket
 
 from ...const import API_PATH, urljoin
-from ...exceptions import APIException
+from ...exceptions import APIException, ClientException
 from ..util import permissions_string, stream_generator
 from ..listing.generator import ListingGenerator
 from ..listing.mixins import SubredditListingMixin
 from .base import RedditBase
+from .emoji import SubredditEmoji
 from .mixins import MessageableMixin
 from .modmail import ModmailConversation
+from .widgets import SubredditWidgets
 from .wikipage import WikiPage
 
 
@@ -52,6 +56,32 @@ class Subreddit(RedditBase, MessageableMixin, SubredditListingMixin):
        for submission in reddit.subreddit('all-redditdev').new():
            print(submission)
 
+    **Typical Attributes**
+
+    This table describes attributes that typically belong to objects of this
+    class. Since attributes are dynamically provided (see
+    :ref:`determine-available-attributes-of-an-object`), there is not a
+    guarantee that these attributes will always be present, nor is this list
+    comprehensive in any way.
+
+    ======================= ===================================================
+    Attribute               Description
+    ======================= ===================================================
+    ``created_utc``         Time the subreddit was created, represented in
+                            `Unix Time`_.
+    ``description``         Subreddit description, in Markdown.
+    ``description_html``    Subreddit description, in HTML.
+    ``display_name``        Name of the subreddit.
+    ``over18``              Whether or not the subreddit is NSFW.
+    ``public_description``  Description of the subreddit, shown in searches
+                            and on the "You must be invited to visit this
+                            community" page (if applicable).
+    ``subscribers``         Count of subscribers to the subreddit.
+    ======================= ===================================================
+
+
+    .. _Unix Time: https://en.wikipedia.org/wiki/Unix_time
+
     """
 
     # pylint: disable=too-many-public-methods
@@ -60,7 +90,8 @@ class Subreddit(RedditBase, MessageableMixin, SubredditListingMixin):
     MESSAGE_PREFIX = '#'
 
     @staticmethod
-    def _create_or_update(_reddit, allow_images=None, allow_top=None,
+    def _create_or_update(_reddit, allow_images=None,
+                          allow_post_crossposts=None, allow_top=None,
                           collapse_deleted_comments=None,
                           comment_score_hide_mins=None, description=None,
                           domain=None, exclude_banned_modqueue=None,
@@ -76,8 +107,9 @@ class Subreddit(RedditBase, MessageableMixin, SubredditListingMixin):
                           suggested_comment_sort=None, title=None,
                           wiki_edit_age=None, wiki_edit_karma=None,
                           wikimode=None, **other_settings):
-        # pylint: disable=invalid-name,too-many-locals
+        # pylint: disable=invalid-name,too-many-locals,too-many-arguments
         model = {'allow_images': allow_images,
+                 'allow_post_crossposts': allow_post_crossposts,
                  'allow_top': allow_top,
                  'collapse_deleted_comments': collapse_deleted_comments,
                  'comment_score_hide_mins': comment_score_hide_mins,
@@ -160,6 +192,31 @@ class Subreddit(RedditBase, MessageableMixin, SubredditListingMixin):
             self._contributor = ContributorRelationship(self, 'contributor')
         return self._contributor
 
+    @property
+    def emoji(self):
+        """Provide an instance of :class:`.SubredditEmoji`.
+
+        This attribute can be used to discover all emoji for a subreddit:
+
+        .. code:: python
+
+           for emoji in reddit.subreddit('iama').emoji:
+               print(emoji)
+
+        A single emoji can be lazily retrieved via:
+
+        .. code:: python
+
+           reddit.subreddit('blah').emoji['emoji_name']
+
+        .. note:: Attempting to access attributes of an nonexistent emoji will
+           result in a :class:`.ClientException`.
+
+        """
+        if self._emoji is None:
+            self._emoji = SubredditEmoji(self)
+        return self._emoji
+
     @property
     def filters(self):
         """Provide an instance of :class:`.SubredditFilters`."""
@@ -281,6 +338,30 @@ class Subreddit(RedditBase, MessageableMixin, SubredditListingMixin):
             self._stylesheet = SubredditStylesheet(self)
         return self._stylesheet
 
+    @property
+    def widgets(self):
+        """Provide an instance of :class:`.SubredditWidgets`.
+
+        **Example usage**
+
+        Get all sidebar widgets:
+
+        .. code-block:: python
+
+           for widget in reddit.subreddit('redditdev').widgets.sidebar:
+               print(widget)
+
+        Get ID card widget:
+
+        .. code-block:: python
+
+           print(reddit.subreddit('redditdev').widgets.id_card)
+
+        """
+        if self._widgets is None:
+            self._widgets = SubredditWidgets(self)
+        return self._widgets
+
     @property
     def wiki(self):
         """Provide an instance of :class:`.SubredditWiki`.
@@ -321,6 +402,7 @@ class Subreddit(RedditBase, MessageableMixin, SubredditListingMixin):
         if display_name:
             self.display_name = display_name
         self._banned = self._contributor = self._filters = self._flair = None
+        self._emoji = self._widgets = None
         self._mod = self._moderator = self._modmail = self._muted = None
         self._quarantine = self._stream = self._stylesheet = self._wiki = None
         self._path = API_PATH['subreddit'].format(subreddit=self)
@@ -328,6 +410,79 @@ class Subreddit(RedditBase, MessageableMixin, SubredditListingMixin):
     def _info_path(self):
         return API_PATH['subreddit_about'].format(subreddit=self)
 
+    def _submit_media(self, data):
+        """Submit and return an `image`, `video`, or `videogif`.
+
+        This is a helper method for submitting posts that are not link posts or
+        self posts.
+        """
+        response = self._reddit.post(API_PATH['submit'], data=data)
+
+        # About the websockets:
+        #
+        # Reddit responds to this request with only two fields: a link to
+        # the user's /submitted page, and a websocket URL. We can use the
+        # websocket URL to get a link to the new post once it is created.
+        #
+        # An important note to PRAW contributors or anyone who would
+        # wish to step through this section with a debugger: This block
+        # of code is NOT debugger-friendly. If there is *any*
+        # significant time between the POST request just above this
+        # comment and the creation of the websocket connection just
+        # below, the code will become stuck in an infinite loop at the
+        # socket.recv() call. I believe this is because only one message is
+        # sent over the websocket, and if the client doesn't connect
+        # soon enough, it will miss the message and get stuck forever
+        # waiting for another.
+        #
+        # So if you need to debug this section of code, please let the
+        # websocket creation happen right after the POST request,
+        # otherwise you will have trouble.
+
+        if not isinstance(response, dict):
+            raise ClientException('Something went wrong with your post: '
+                                  '{!r}'.format(response))
+
+        try:
+            socket = websocket.create_connection(response['json']['data']
+                                                 ['websocket_url'],
+                                                 timeout=2)
+            ws_update = loads(socket.recv())
+            socket.close(timeout=2)
+        except websocket.WebSocketTimeoutException:
+            raise ClientException('Websocket error. Check your media file. '
+                                  'Your post may still have been created.')
+        url = ws_update['payload']['redirect']
+        return self._reddit.submission(url=url)
+
+    def _upload_media(self, media_path):
+        """Upload media and return its URL. Uses undocumented endpoint."""
+        if media_path is None:
+            media_path = join(dirname(dirname(dirname(__file__))),
+                              'images', 'PRAW logo.png')
+
+        file_name = basename(media_path).lower()
+        mime_type = {'png': 'image/png', 'mov': 'video/quicktime',
+                     'mp4': 'video/mp4', 'jpg': 'image/jpeg',
+                     'jpeg': 'image/jpeg',
+                     'gif': 'image/gif'}.get(file_name.rpartition('.')[2],
+                                             'image/jpeg')  # default to JPEG
+        img_data = {'filepath': file_name, 'mimetype': mime_type}
+
+        url = API_PATH['media_asset']
+        # until we learn otherwise, assume this request always succeeds
+        upload_lease = self._reddit.post(url, data=img_data)['args']
+        upload_url = 'https:{}'.format(upload_lease['action'])
+        upload_data = {item['name']: item['value']
+                       for item in upload_lease['fields']}
+
+        with open(media_path, 'rb') as media:
+            response = self._reddit._core._requestor._http.post(
+                upload_url, data=upload_data, files={'file': media})
+        response.raise_for_status()
+
+        return upload_url + '/' + upload_data['key']
+
     def random(self):
         """Return a random Submission."""
         url = API_PATH['subreddit_random'].format(subreddit=self)
@@ -398,65 +553,6 @@ class Subreddit(RedditBase, MessageableMixin, SubredditListingMixin):
         return self._submission_class(self._reddit, url=urljoin(
             self._reddit.config.reddit_url, path))
 
-    def submissions(self, start=None, end=None, extra_query=None):
-        """Yield submissions created between timestamps ``start`` and ``end``.
-
-        :param start: A UNIX timestamp indicating the earliest creation time of
-            submission yielded during the call. A value of ``None`` will
-            consider all submissions older than ``end`` (default: None).
-        :param end: A UNIX timestamp indicating the latest creation time of a
-            submission yielded during the call. A value of ``None`` will
-            consider all submissions newer than ``start`` (default: None).
-        :param extra_query: A cloudsearch query that will be combined via
-            ``(and timestamp:start..end EXTRA_QUERY)`` to further filter
-            results (default: None).
-
-        Submissions are yielded newest first.
-
-        **Example**: Suppose you want to obtain all submissions to
-        ``/r/politics`` on November 8, 2016 PST. First, you need to determine
-        the start and end dates as UNIX timestamps. Using
-        http://www.epochconverter.com/, those timestamps are 1478592000 and
-        1478678400, respectively. The following outputs all such submissions'
-        titles:
-
-        .. code:: python
-
-           subreddit = reddit.subreddit('politics')
-           for submission in subreddit.submissions(1478592000, 1478678400):
-               print(submission.title)
-
-        As of this writing there are 809 results.
-
-        .. note:: The results are only as reliable as reddit's search.
-           Submissions may be missing from the results.
-
-        """
-        utc_offset = 28800
-        now = int(time.time())
-        start = max(int(start) + utc_offset if start else 0, 0)
-        end = min(int(end) if end else now, now) + utc_offset
-
-        found_new_submission = True
-        last_ids = set()
-        params = {}
-        while found_new_submission:
-            query = 'timestamp:{}..{}'.format(start, end)
-            if extra_query:
-                query = '(and {} {})'.format(query, extra_query)
-
-            current_ids = set()
-            found_new_submission = False
-            for submission in self.search(query, limit=None, params=params,
-                                          sort='new', syntax='cloudsearch'):
-                current_ids.add(submission.id)
-                end = min(end, int(submission.created))
-                if submission.id not in last_ids:
-                    found_new_submission = True
-                yield submission
-                params['after'] = submission.fullname
-            last_ids = current_ids
-
     def submit(self, title, selftext=None, url=None, flair_id=None,
                flair_text=None, resubmit=True, send_replies=True):
         """Add a submission to the subreddit.
@@ -486,6 +582,11 @@ class Subreddit(RedditBase, MessageableMixin, SubredditListingMixin):
            url = 'https://praw.readthedocs.io'
            reddit.subreddit('reddit_api_test').submit(title, url=url)
 
+        .. note ::
+
+           For submitting images, videos, and videogifs,
+           see :meth:`.submit_image` and :meth:`.submit_video`.
+
         """
         if (bool(selftext) or selftext == '') == bool(url):
             raise TypeError('Either `selftext` or `url` must be provided.')
@@ -501,6 +602,99 @@ class Subreddit(RedditBase, MessageableMixin, SubredditListingMixin):
             data.update(kind='link', url=url)
         return self._reddit.post(API_PATH['submit'], data=data)
 
+    def submit_image(self, title, image_path, flair_id=None,
+                     flair_text=None, resubmit=True, send_replies=True):
+        """Add an image submission to the subreddit.
+
+        :param title: The title of the submission.
+        :param image_path: The path to an image, to upload and post.
+        :param flair_id: The flair template to select (default: None).
+        :param flair_text: If the template's ``flair_text_editable`` value is
+            True, this value will set a custom text (default: None).
+        :param resubmit: When False, an error will occur if the URL has already
+            been submitted (default: True).
+        :param send_replies: When True, messages will be sent to the submission
+            author when comments are made to the submission (default: True).
+
+        .. note::
+
+           Reddit's API uses WebSockets to respond with the link of the
+           newly created post. If this fails, the method will raise
+           :class:`.ClientException`. Occasionally, the Reddit post will still
+           be created. More often, there is an error with the image file.
+
+        :returns: A :class:`~.Submission` object for the newly created
+            submission.
+
+        For example to submit an image to ``/r/reddit_api_test`` do:
+
+        .. code:: python
+
+           title = 'My favorite picture'
+           image = '/path/to/image.png'
+           reddit.subreddit('reddit_api_test').submit_image(title, image)
+
+        """
+        data = {'sr': str(self), 'resubmit': bool(resubmit),
+                'sendreplies': bool(send_replies), 'title': title}
+        for key, value in (('flair_id', flair_id), ('flair_text', flair_text)):
+            if value is not None:
+                data[key] = value
+        data.update(kind='image', url=self._upload_media(image_path))
+        return self._submit_media(data)
+
+    def submit_video(self, title, video_path, videogif=False,
+                     thumbnail_path=None, flair_id=None, flair_text=None,
+                     resubmit=True, send_replies=True):
+        """Add a video or videogif submission to the subreddit.
+
+        :param title: The title of the submission.
+        :param video_path: The path to a video, to upload and post.
+        :param videogif: A ``bool`` value. If ``True``, the video is
+            uploaded as a videogif, which is essentially a silent video (
+            default: ``False``).
+        :param thumbnail_path: (Optional) The path to an image, to be uploaded
+            and used as the thumbnail for this video. If not provided, the
+            PRAW logo will be used as the thumbnail.
+        :param flair_id: The flair template to select (default: ``None``).
+        :param flair_text: If the template's ``flair_text_editable`` value is
+            True, this value will set a custom text (default: ``None``).
+        :param resubmit: When False, an error will occur if the URL has already
+            been submitted (default: ``True``).
+        :param send_replies: When True, messages will be sent to the submission
+            author when comments are made to the submission
+            (default: ``True``).
+
+        .. note::
+
+           Reddit's API uses WebSockets to respond with the link of the
+           newly created post. If this fails, the method will raise
+           :class:`.ClientException`. Occasionally, the Reddit post will still
+           be created. More often, there is an error with the video file.
+
+        :returns: A :class:`~.Submission` object for the newly created
+            submission.
+
+        For example to submit a video to ``/r/reddit_api_test`` do:
+
+        .. code:: python
+
+           title = 'My favorite movie'
+           video = '/path/to/video.mp4'
+           reddit.subreddit('reddit_api_test').submit_video(title, video)
+
+        """
+        data = {'sr': str(self), 'resubmit': bool(resubmit),
+                'sendreplies': bool(send_replies), 'title': title}
+        for key, value in (('flair_id', flair_id), ('flair_text', flair_text)):
+            if value is not None:
+                data[key] = value
+        data.update(kind='videogif' if videogif else 'video',
+                    url=self._upload_media(video_path),
+                    # if thumbnail_path is None, it uploads the PRAW logo
+                    video_poster_url=self._upload_media(thumbnail_path))
+        return self._submit_media(data)
+
     def subscribe(self, other_subreddits=None):
         """Subscribe to the subreddit.
 
@@ -598,7 +792,7 @@ class SubredditFilters(object):
             special=self.subreddit, user=self.subreddit._reddit.user.me(),
             subreddit=subreddit)
         self.subreddit._reddit.request(
-            'PUT', url, data={'model': dumps({'name': subreddit})})
+            'PUT', url, data={'model': dumps({'name': str(subreddit)})})
 
     def remove(self, subreddit):
         """Remove ``subreddit`` from the list of filtered subreddits.
@@ -610,7 +804,7 @@ class SubredditFilters(object):
         """
         url = API_PATH['subreddit_filter'].format(
             special=self.subreddit, user=self.subreddit._reddit.user.me(),
-            subreddit=subreddit)
+            subreddit=str(subreddit))
         self.subreddit._reddit.request('DELETE', url, data={})
 
 
@@ -732,7 +926,8 @@ class SubredditFlair(object):
         """
         return self.update(x['user'] for x in self())
 
-    def set(self, redditor=None, text='', css_class=''):
+    def set(self, redditor=None, text='', css_class='',
+            flair_template_id=None):
         """Set flair for a Redditor.
 
         :param redditor: (Required) A redditor name (e.g., ``'spez'``) or
@@ -740,7 +935,9 @@ class SubredditFlair(object):
         :param text: The flair text to associate with the Redditor or
             Submission (default: '').
         :param css_class: The css class to associate with the flair html
-            (default: '').
+            (default: ''). Use either this or ``flair_template_id``.
+        :param flair_template_id: The ID of the flair template to be used
+            (default: ``None``). Use either this or ``css_class``.
 
         This method can only be used by an authenticated user who is a
         moderator of the associated Subreddit.
@@ -749,11 +946,23 @@ class SubredditFlair(object):
 
         .. code:: python
 
-           reddit.subreddit('redditdev').flair.set('bboe', 'PRAW author')
+           reddit.subreddit('redditdev').flair.set('bboe', 'PRAW author',
+                                                   css_class='mods')
+           template = '6bd28436-1aa7-11e9-9902-0e05ab0fad46'
+           reddit.subreddit('redditdev').flair.set('spez', 'Reddit CEO',
+                                                   flair_template_id=template)
 
         """
-        data = {'css_class': css_class, 'name': str(redditor), 'text': text}
-        url = API_PATH['flair'].format(subreddit=self.subreddit)
+        if css_class and flair_template_id is not None:
+            raise TypeError('Parameter `css_class` cannot be used in '
+                            'conjunction with `flair_template_id`.')
+        data = {'name': str(redditor), 'text': text}
+        if flair_template_id is not None:
+            data['flair_template_id'] = flair_template_id
+            url = API_PATH['select_flair'].format(subreddit=self.subreddit)
+        else:
+            data['css_class'] = css_class
+            url = API_PATH['flair'].format(subreddit=self.subreddit)
         self.subreddit._reddit.post(url, data=data)
 
     def update(self, flair_list, text='', css_class=''):
@@ -823,10 +1032,25 @@ class SubredditFlairTemplates(object):
         """
         self.subreddit = subreddit
 
-    def _add(self, text, css_class='', text_editable=False, is_link=None):
-        url = API_PATH['flairtemplate'].format(subreddit=self.subreddit)
-        data = {'css_class': css_class, 'flair_type': self.flair_type(is_link),
-                'text': text, 'text_editable': bool(text_editable)}
+    def _add(self, text, css_class='', text_editable=False, is_link=None,
+             background_color=None, text_color=None, mod_only=None):
+        if css_class and any(param is not None for param in
+                             (background_color, text_color, mod_only)):
+            raise TypeError('Parameter `css_class` cannot be used in '
+                            'conjunction with parameters `background_color`, '
+                            '`text_color`, or `mod_only`.')
+        if css_class:
+            url = API_PATH['flairtemplate'].format(subreddit=self.subreddit)
+            data = {'css_class': css_class,
+                    'flair_type': self.flair_type(is_link),
+                    'text': text, 'text_editable': bool(text_editable)}
+        else:
+            url = API_PATH['flairtemplate_v2'].format(subreddit=self.subreddit)
+            data = {'background_color': background_color,
+                    'text_color': text_color,
+                    'flair_type': self.flair_type(is_link),
+                    'text': text, 'text_editable': bool(text_editable),
+                    'mod_only': bool(mod_only)}
         self.subreddit._reddit.post(url, data=data)
 
     def _clear(self, is_link=None):
@@ -842,35 +1066,64 @@ class SubredditFlairTemplates(object):
         .. code-block:: python
 
            template_info = list(subreddit.flair.templates)[0]
-           subreddit.flair.templates.delete(template_info['flair_template_id])
+           subreddit.flair.templates.delete(template_info['id'])
 
         """
         url = API_PATH['flairtemplatedelete'].format(subreddit=self.subreddit)
         self.subreddit._reddit.post(
             url, data={'flair_template_id': template_id})
 
-    def update(self, template_id, text, css_class='', text_editable=False):
-        """Update the flair templated provided by ``template_id``.
+    def update(self, template_id, text, css_class='', text_editable=False,
+               background_color=None, text_color=None, mod_only=None):
+        """Update the flair template provided by ``template_id``.
 
         :param template_id: The flair template to update.
         :param text: The flair template's new text (required).
         :param css_class: The flair template's new css_class (default: '').
+            Cannot be used in conjunction with ``background_color``,
+            ``text_color``, or ``mod_only``.
         :param text_editable: (boolean) Indicate if the flair text can be
             modified for each Redditor that sets it (default: False).
+        :param background_color: The flair template's new background color,
+            as a hex color. Cannot be used in conjunction with ``css_class``.
+        :param text_color: The flair template's new text color, either
+            ``'light'`` or ``'dark'``. Cannot be used in conjunction with
+            ``css_class``.
+        :param mod_only: (boolean) Indicate if the flair can only be used by
+            moderators. Cannot be used in conjunction with ``css_class``.
 
-        For example to make a link flair template text_editable, try:
+        For example to make a user flair template text_editable, try:
 
         .. code-block:: python
 
            template_info = list(subreddit.flair.templates)[0]
            subreddit.flair.templates.update(
-               template_info['flair_template_id'],
+               template_info['id'],
+               template_info['flair_text'],
                text_editable=True)
 
+        .. note::
+
+           Any parameters not provided will be set to default values (usually
+           ``None`` or ``False``) on Reddit's end.
+
         """
-        url = API_PATH['flairtemplate'].format(subreddit=self.subreddit)
-        data = {'css_class': css_class, 'flair_template_id': template_id,
-                'text': text, 'text_editable': bool(text_editable)}
+        if css_class and any(param is not None for param in
+                             (background_color, text_color, mod_only)):
+            raise TypeError('Parameter `css_class` cannot be used in '
+                            'conjunction with parameters `background_color`, '
+                            '`text_color`, or `mod_only`.')
+
+        if css_class:
+            url = API_PATH['flairtemplate'].format(subreddit=self.subreddit)
+            data = {'css_class': css_class, 'flair_template_id': template_id,
+                    'text': text, 'text_editable': bool(text_editable)}
+        else:
+            url = API_PATH['flairtemplate_v2'].format(subreddit=self.subreddit)
+            data = {'flair_template_id': template_id, 'text': text,
+                    'background_color': background_color,
+                    'text_color': text_color, 'text_editable': text_editable,
+                    'mod_only': mod_only}
         self.subreddit._reddit.post(url, data=data)
 
 
@@ -889,18 +1142,28 @@ class SubredditRedditorFlairTemplates(SubredditFlairTemplates):
 
 
         """
-        url = API_PATH['flairselector'].format(subreddit=self.subreddit)
-        data = {'unique': self.subreddit._reddit._next_unique}
-        for template in self.subreddit._reddit.post(url, data=data)['choices']:
+        url = API_PATH['user_flair'].format(subreddit=self.subreddit)
+        params = {'unique': self.subreddit._reddit._next_unique}
+        for template in self.subreddit._reddit.get(url, params=params):
             yield template
 
-    def add(self, text, css_class='', text_editable=False):
+    def add(self, text, css_class='', text_editable=False,
+            background_color=None, text_color=None, mod_only=None):
         """Add a Redditor flair template to the associated subreddit.
 
         :param text: The flair template's text (required).
         :param css_class: The flair template's css_class (default: '').
+            Cannot be used in conjunction with ``background_color``,
+            ``text_color``, or ``mod_only``.
         :param text_editable: (boolean) Indicate if the flair text can be
             modified for each Redditor that sets it (default: False).
+        :param background_color: The flair template's new background color,
+            as a hex color. Cannot be used in conjunction with ``css_class``.
+        :param text_color: The flair template's new text color, either
+            ``'light'`` or ``'dark'``. Cannot be used in conjunction with
+            ``css_class``.
+        :param mod_only: (boolean) Indicate if the flair can only be used by
+            moderators. Cannot be used in conjunction with ``css_class``.
 
         For example, to add an editable Redditor flair try:
 
@@ -911,7 +1174,8 @@ class SubredditRedditorFlairTemplates(SubredditFlairTemplates):
 
         """
         self._add(text, css_class=css_class, text_editable=text_editable,
-                  is_link=False)
+                  is_link=False, background_color=background_color,
+                  text_color=text_color, mod_only=mod_only)
 
     def clear(self):
         """Remove all Redditor flair templates from the subreddit.
@@ -945,13 +1209,23 @@ class SubredditLinkFlairTemplates(SubredditFlairTemplates):
         for template in self.subreddit._reddit.get(url):
             yield template
 
-    def add(self, text, css_class='', text_editable=False):
+    def add(self, text, css_class='', text_editable=False,
+            background_color=None, text_color=None, mod_only=None):
         """Add a link flair template to the associated subreddit.
 
         :param text: The flair template's text (required).
         :param css_class: The flair template's css_class (default: '').
+            Cannot be used in conjunction with ``background_color``,
+            ``text_color``, or ``mod_only``.
         :param text_editable: (boolean) Indicate if the flair text can be
             modified for each Redditor that sets it (default: False).
+        :param background_color: The flair template's new background color,
+            as a hex color. Cannot be used in conjunction with ``css_class``.
+        :param text_color: The flair template's new text color, either
+            ``'light'`` or ``'dark'``. Cannot be used in conjunction with
+            ``css_class``.
+        :param mod_only: (boolean) Indicate if the flair can only be used by
+            moderators. Cannot be used in conjunction with ``css_class``.
 
         For example, to add an editable link flair try:
 
@@ -962,7 +1236,8 @@ class SubredditLinkFlairTemplates(SubredditFlairTemplates):
 
         """
         self._add(text, css_class=css_class, text_editable=text_editable,
-                  is_link=True)
+                  is_link=True, background_color=background_color,
+                  text_color=text_color, mod_only=mod_only)
 
     def clear(self):
         """Remove all link flair templates from the subreddit.
@@ -1183,6 +1458,8 @@ class SubredditModeration(object):
 
         :param allow_images: Allow users to upload images using the native
             image hosting. Only applies to link-only subreddits.
+        :param allow_post_crossposts: Allow users to crosspost submissions from
+            other subreddits.
         :param allow_top: Allow the subreddit to appear on ``/r/all`` as well
             as the default and trending lists.
         :param collapse_deleted_comments: Collapse deleted and removed comments
@@ -1484,6 +1761,7 @@ class ModeratorRelationship(SubredditRelationship):
         """
         other_settings = self._handle_permissions(permissions, other_settings)
         super(ModeratorRelationship, self).add(redditor, **other_settings)
+
     # pylint: enable=arguments-differ
 
     def invite(self, redditor, permissions=None, **other_settings):
@@ -1519,8 +1797,7 @@ class ModeratorRelationship(SubredditRelationship):
            reddit.subreddit('subredditname').moderator.leave()
 
         """
-        self.subreddit._reddit.post(API_PATH['leavemoderator'],
-                                    data={'id': self.subreddit.fullname})
+        self.remove(self.subreddit._reddit.config.username)
 
     def remove_invite(self, redditor):
         """Remove the moderator invite for ``redditor``.
@@ -1594,7 +1871,7 @@ class ModeratorRelationship(SubredditRelationship):
 class Modmail(object):
     """Provides modmail functions for a subreddit."""
 
-    def __call__(self, id=None, mark_read=False):
+    def __call__(self, id=None, mark_read=False):  # noqa: D207, D301
         """Return an individual conversation.
 
         :param id: A reddit base36 conversation ID, e.g., ``2gmz``.
@@ -1611,8 +1888,8 @@ class Modmail(object):
 
         .. code:: python
 
-           conversation = reddit.subreddit('redditdev').modmail('2gmz',
-                                                                mark_read=True)
+           conversation = reddit.subreddit('redditdev').modmail('2gmz', \
+mark_read=True)
            for message in conversation.messages:
                print(message.body_markdown)
 
@@ -1626,16 +1903,16 @@ class Modmail(object):
 
         .. code:: python
 
-           conversation = reddit.subreddit('redditdev').modmail('2gmz',
-                                                                mark_read=True)
+           conversation = reddit.subreddit('redditdev').modmail('2gmz', \
+mark_read=True)
            print(conversation.user.ban_status)
 
         To print a list of recent submissions by the user:
 
         .. code:: python
 
-           conversation = reddit.subreddit('redditdev').modmail('2gmz',
-                                                                mark_read=True)
+           conversation = reddit.subreddit('redditdev').modmail('2gmz', \
+mark_read=True)
            print(conversation.user.recent_posts)
 
         """
@@ -1684,7 +1961,7 @@ class Modmail(object):
                 for conversation_id in response['conversation_ids']]
 
     def conversations(self, after=None, limit=None, other_subreddits=None,
-                      sort=None, state=None):
+                      sort=None, state=None):  # noqa: D207, D301
         """Generate :class:`.ModmailConversation` objects for subreddit(s).
 
         :param after: A base36 modmail conversation id. When provided, the
@@ -1705,7 +1982,8 @@ class Modmail(object):
 
         .. code:: python
 
-            conversations = reddit.subreddit('all').conversations(state='mod')
+            conversations = reddit.subreddit('all').modmail.conversations(\
+state='mod')
 
         """
         params = {}
@@ -1807,7 +2085,7 @@ class SubredditStream(object):
         Comments are yielded oldest first. Up to 100 historical comments will
         initially be returned.
 
-        Keyword arguments are passed to :meth:`.stream_generator`.
+        Keyword arguments are passed to :func:`.stream_generator`.
 
         For example, to retrieve all new comments made to the ``iama``
         subreddit, try:
@@ -1817,6 +2095,15 @@ class SubredditStream(object):
            for comment in reddit.subreddit('iama').stream.comments():
                print(comment)
 
+        To only retreive new submissions starting when the stream is
+        created, pass `skip_existing=True`:
+
+        .. code:: python
+
+           subreddit = reddit.subreddit('iama')
+           for comment in subreddit.stream.comments(skip_existing=True):
+               print(comment)
+
         """
         return stream_generator(self.subreddit.comments, **stream_options)
 
@@ -1826,7 +2113,7 @@ class SubredditStream(object):
         Submissions are yielded oldest first. Up to 100 historical submissions
         will initially be returned.
 
-        Keyword arguments are passed to :meth:`.stream_generator`.
+        Keyword arguments are passed to :func:`.stream_generator`.
 
         For example to retrieve all new submissions made to all of Reddit, try:
 
