@@ -1,8 +1,9 @@
 """Test praw.models.subreddit."""
 from os.path import abspath, dirname, join
+from json import dumps
 import sys
 
-from praw.exceptions import APIException
+from praw.exceptions import APIException, ClientException
 from praw.models import (Comment, ModAction, ModmailAction,
                          ModmailConversation, ModmailMessage, Redditor,
                          Submission, Subreddit, SubredditMessage, Stylesheet,
@@ -10,11 +11,39 @@ from praw.models import (Comment, ModAction, ModmailAction,
 from prawcore import Forbidden, NotFound, RequestException, TooLarge
 import mock
 import pytest
+import websocket
 
 from ... import IntegrationTest
 
 
+class WebsocketMock(object):
+    POST_URL = 'https://reddit.com/r/<TEST_SUBREDDIT>/comments/{}/test_title/'
+
+    @classmethod
+    def make_dict(cls, post_id):
+        return {'payload': {'redirect': cls.POST_URL.format(post_id)}}
+
+    def __init__(self, *post_ids):
+        self.post_ids = post_ids
+        self.i = -1
+
+    def close(self, *args, **kwargs):
+        pass
+
+    def recv(self):
+        if not self.post_ids:
+            raise websocket.WebSocketTimeoutException()
+        assert 0 <= self.i + 1 < len(self.post_ids)
+        self.i += 1
+        return dumps(self.make_dict(self.post_ids[self.i]))
+
+
 class TestSubreddit(IntegrationTest):
+    @staticmethod
+    def image_path(name):
+        test_dir = abspath(dirname(sys.modules[__name__].__file__))
+        return join(test_dir, '..', '..', 'files', name)
+
     @mock.patch('time.sleep', return_value=None)
     def test_create(self, _):
         self.reddit.read_only = False
@@ -153,6 +182,172 @@ class TestSubreddit(IntegrationTest):
             assert submission.url == url
             assert submission.title == 'Test Title'
 
+    @mock.patch('time.sleep', return_value=None)
+    @mock.patch('websocket.create_connection',
+                return_value=WebsocketMock('ahf0uh',  # update with cassette
+                                           'ahf0uq',  # update with cassette
+                                           'ahf0v4'))  # update with cassette
+    def test_submit_image(self, _, __):
+        self.reddit.read_only = False
+        with self.recorder.use_cassette('TestSubreddit.test_submit_image'):
+            subreddit = self.reddit.subreddit(
+                pytest.placeholders.test_subreddit)
+            for file_name in ('test.png', 'test.jpg', 'test.gif'):
+                image = self.image_path(file_name)
+
+                submission = subreddit.submit_image('Test Title', image)
+                assert submission.author == self.reddit.config.username
+                assert submission.is_reddit_media_domain
+                assert submission.title == 'Test Title'
+
+    @mock.patch('time.sleep', return_value=None)
+    @mock.patch('websocket.create_connection', return_value=WebsocketMock())
+    def test_submit_image__bad_websocket(self, _, __):
+        self.reddit.read_only = False
+        with self.recorder.use_cassette('TestSubreddit.test_submit_image'):
+            subreddit = self.reddit.subreddit(
+                pytest.placeholders.test_subreddit)
+            for file_name in ('test.png', 'test.jpg'):
+                image = self.image_path(file_name)
+
+                with pytest.raises(ClientException):
+                    subreddit.submit_image('Test Title', image)
+
+    @mock.patch('time.sleep', return_value=None)
+    def test_submit_image__bad_filetype(self, _):
+        self.reddit.read_only = False
+        with self.recorder.use_cassette(
+                'TestSubreddit.test_submit_image__bad_filetype'):
+            subreddit = self.reddit.subreddit(
+                pytest.placeholders.test_subreddit)
+
+            image = self.image_path('test.mov')
+            with pytest.raises(ClientException):
+                subreddit.submit_image('Test Title', image)
+
+    @mock.patch('time.sleep', return_value=None)
+    @mock.patch('websocket.create_connection',
+                return_value=WebsocketMock('ah3gqo'))  # update with cassette
+    def test_submit_image__flair(self, _, __):
+        flair_id = '6bd28436-1aa7-11e9-9902-0e05ab0fad46'
+        flair_text = 'Test flair text'
+        flair_class = 'test-flair-class'
+        self.reddit.read_only = False
+        with self.recorder.use_cassette(
+                'TestSubreddit.test_submit_image__flair'):
+            subreddit = self.reddit.subreddit(
+                pytest.placeholders.test_subreddit)
+            image = self.image_path('test.jpg')
+            submission = subreddit.submit_image('Test Title', image,
+                                                flair_id=flair_id,
+                                                flair_text=flair_text)
+            assert submission.link_flair_css_class == flair_class
+            assert submission.link_flair_text == flair_text
+
+    @mock.patch('time.sleep', return_value=None)
+    @mock.patch('websocket.create_connection',
+                return_value=WebsocketMock('aheljy',  # update with cassette
+                                           'ahelks'))  # update with cassette
+    def test_submit_video(self, _, __):
+        self.reddit.read_only = False
+        with self.recorder.use_cassette('TestSubreddit.test_submit_video'):
+            subreddit = self.reddit.subreddit(
+                pytest.placeholders.test_subreddit)
+            for file_name in ('test.mov', 'test.mp4'):
+                video = self.image_path(file_name)
+
+                submission = subreddit.submit_video('Test Title', video)
+                assert submission.author == self.reddit.config.username
+                assert submission.is_video
+                assert submission.title == 'Test Title'
+
+    @mock.patch('time.sleep', return_value=None)
+    @mock.patch('websocket.create_connection', return_value=WebsocketMock())
+    def test_submit_video__bad_filetype(self, _, __):
+        self.reddit.read_only = False
+        with self.recorder.use_cassette(
+                'TestSubreddit.test_submit_video__bad_filetype'):
+            subreddit = self.reddit.subreddit(
+                pytest.placeholders.test_subreddit)
+            for file_name in ('test.mov', 'test.mp4'):
+                video = self.image_path(file_name)
+
+                with pytest.raises(ClientException):
+                    subreddit.submit_video('Test Title', video,
+                                           thumbnail_path=video)
+
+    @mock.patch('time.sleep', return_value=None)
+    @mock.patch('websocket.create_connection', return_value=WebsocketMock())
+    def test_submit_video__bad_websocket(self, _, __):
+        self.reddit.read_only = False
+        with self.recorder.use_cassette('TestSubreddit.test_submit_video'):
+            subreddit = self.reddit.subreddit(
+                pytest.placeholders.test_subreddit)
+            for file_name in ('test.mov', 'test.mp4'):
+                video = self.image_path(file_name)
+
+                with pytest.raises(ClientException):
+                    subreddit.submit_video('Test Title', video)
+
+    @mock.patch('time.sleep', return_value=None)
+    @mock.patch('websocket.create_connection',
+                return_value=WebsocketMock('ahells'))  # update with cassette
+    def test_submit_video__flair(self, _, __):
+        flair_id = '6bd28436-1aa7-11e9-9902-0e05ab0fad46'
+        flair_text = 'Test flair text'
+        flair_class = 'test-flair-class'
+        self.reddit.read_only = False
+        with self.recorder.use_cassette(
+                'TestSubreddit.test_submit_video__flair'):
+            subreddit = self.reddit.subreddit(
+                pytest.placeholders.test_subreddit)
+            image = self.image_path('test.mov')
+            submission = subreddit.submit_video('Test Title', image,
+                                                flair_id=flair_id,
+                                                flair_text=flair_text)
+            assert submission.link_flair_css_class == flair_class
+            assert submission.link_flair_text == flair_text
+
+    @mock.patch('time.sleep', return_value=None)
+    @mock.patch('websocket.create_connection',
+                return_value=WebsocketMock('aheln2',  # update with cassette
+                                           'ahelnz'))  # update with cassette
+    def test_submit_video__thumbnail(self, _, __):
+        self.reddit.read_only = False
+        with self.recorder.use_cassette(
+                'TestSubreddit.test_submit_video__thumbnail'):
+            subreddit = self.reddit.subreddit(
+                pytest.placeholders.test_subreddit)
+            for video_name, thumb_name in (('test.mov', 'test.jpg'),
+                                           ('test.mp4', 'test.png')):
+                video = self.image_path(video_name)
+                thumb = self.image_path(thumb_name)
+
+                submission = subreddit.submit_video('Test Title', video,
+                                                    thumbnail_path=thumb)
+                assert submission.author == self.reddit.config.username
+                assert submission.is_video
+                assert submission.title == 'Test Title'
+
+    @mock.patch('time.sleep', return_value=None)
+    @mock.patch('websocket.create_connection',
+                return_value=WebsocketMock('ahelor',  # update with cassette
+                                           'ahelpf'))  # update with cassette
+    def test_submit_video__videogif(self, _, __):
+        self.reddit.read_only = False
+        with self.recorder.use_cassette(
+                'TestSubreddit.test_submit_video__videogif'):
+            subreddit = self.reddit.subreddit(
+                pytest.placeholders.test_subreddit)
+            for file_name in ('test.mov', 'test.mp4'):
+                video = self.image_path(file_name)
+
+                submission = subreddit.submit_video('Test Title', video,
+                                                    videogif=True)
+                assert submission.author == self.reddit.config.username
+                assert submission.is_video
+                assert submission.title == 'Test Title'
+
     def test_subscribe(self):
         self.reddit.read_only = False
         subreddit = self.reddit.subreddit(pytest.placeholders.test_subreddit)
@@ -303,6 +498,23 @@ class TestSubredditFlair(IntegrationTest):
             assert len(response) > 100
             assert all('removed' in x['status'] for x in response)
 
+    def test_set__flair_id(self):
+        self.reddit.read_only = False
+        with self.recorder.use_cassette(
+                'TestSubredditFlair.test_set__flair_id'):
+            redditor = self.reddit.redditor(self.reddit.config.username)
+            flair = '11c32eee-1482-11e9-bfc0-0efc81a5e8e8'
+            self.subreddit.flair.set(redditor, 'redditor flair',
+                                     flair_template_id=flair)
+
+    def test_set__flair_id_default_text(self):
+        self.reddit.read_only = False
+        with self.recorder.use_cassette(
+                'TestSubredditFlair.test_set__flair_id_default_text'):
+            redditor = self.reddit.redditor(self.reddit.config.username)
+            flair = '11c32eee-1482-11e9-bfc0-0efc81a5e8e8'
+            self.subreddit.flair.set(redditor, flair_template_id=flair)
+
     def test_set__redditor(self):
         self.reddit.read_only = False
         with self.recorder.use_cassette(
@@ -311,6 +523,13 @@ class TestSubredditFlair(IntegrationTest):
                 self.reddit.config.username)
             self.subreddit.flair.set(redditor, 'redditor flair')
 
+    def test_set__redditor_css_only(self):
+        self.reddit.read_only = False
+        with self.recorder.use_cassette(
+                'TestSubredditFlair.test_set__redditor_css_only'):
+            self.subreddit.flair.set(self.reddit.config.username,
+                                     css_class='some class')
+
     def test_set__redditor_string(self):
         self.reddit.read_only = False
         with self.recorder.use_cassette(
@@ -318,13 +537,6 @@ class TestSubredditFlair(IntegrationTest):
             self.subreddit.flair.set(self.reddit.config.username, 'new flair',
                                      'some class')
 
-    def test_set__submission(self):
-        self.reddit.read_only = False
-        with self.recorder.use_cassette(
-                'TestSubredditFlair.test_set__submission'):
-            submission = self.subreddit._reddit.submission('4b536p')
-            self.subreddit.flair.set(submission, 'submission flair')
-
     def test_update(self):
         self.reddit.read_only = False
         redditor = self.subreddit._reddit.redditor(self.reddit.config.username)
@@ -366,13 +578,26 @@ class TestSubredditFlairTemplates(IntegrationTest):
             templates = list(self.subreddit.flair.templates)
         assert len(templates) > 100
 
+        for flair_template in templates:
+            assert flair_template['id']
+
     @mock.patch('time.sleep', return_value=None)
     def test_add(self, _):
         self.reddit.read_only = False
         with self.recorder.use_cassette(
                 'TestSubredditFlairTemplates.test_add'):
             for i in range(101):
-                self.subreddit.flair.templates.add('PRAW{}'.format(i))
+                self.subreddit.flair.templates.add('PRAW{}'.format(i),
+                                                   css_class='myCSS')
+
+    @mock.patch('time.sleep', return_value=None)
+    def test_add_v2(self, _):
+        self.reddit.read_only = False
+        with self.recorder.use_cassette(
+                'TestSubredditFlairTemplates.test_add_v2'):
+            for i in range(101):
+                self.subreddit.flair.templates.add(
+                    'PRAW{}'.format(i), background_color='#ABCDEF')
 
     def test_clear(self):
         self.reddit.read_only = False
@@ -387,7 +612,7 @@ class TestSubredditFlairTemplates(IntegrationTest):
                 'TestSubredditFlairTemplates.test_delete'):
             template = list(self.subreddit.flair.templates)[0]
             self.subreddit.flair.templates.delete(
-                template['flair_template_id'])
+                template['id'])
 
     @mock.patch('time.sleep', return_value=None)
     def test_update(self, _):
@@ -396,7 +621,40 @@ class TestSubredditFlairTemplates(IntegrationTest):
                 'TestSubredditFlairTemplates.test_update'):
             template = list(self.subreddit.flair.templates)[0]
             self.subreddit.flair.templates.update(
-                template['flair_template_id'], 'PRAW updated')
+                template['id'], 'PRAW updated', css_class='myCSS')
+
+    def test_update_bad(self):
+        self.reddit.read_only = False
+        with self.recorder.use_cassette(
+                'TestSubredditFlairTemplates.test_update_bad'):
+            template = list(self.subreddit.flair.templates)[0]
+            with pytest.raises(TypeError):
+                self.subreddit.flair.templates.update(
+                    template['id'], 'PRAW updated', css_class='myclass',
+                    text_color='light')
+
+    @mock.patch('time.sleep', return_value=None)
+    def test_update_across_kind(self, _):
+        self.reddit.read_only = False
+        with self.recorder.use_cassette(
+                'TestSubredditFlairTemplates.test_update_across_kind'):
+            self.subreddit.flair.templates.add('My PRAW text', css_class='flr')
+            template = [t for t in self.subreddit.flair.templates
+                        if t['text'] == 'My PRAW text'][0]
+            self.subreddit.flair.templates.update(template['id'],
+                                                  'updated text :snoo:',
+                                                  text_color='dark',
+                                                  background_color='#00FFFF')
+
+    @mock.patch('time.sleep', return_value=None)
+    def test_update_v2(self, _):
+        self.reddit.read_only = False
+        with self.recorder.use_cassette(
+                'TestSubredditFlairTemplates.test_update_v2'):
+            template = list(self.subreddit.flair.templates)[0]
+            self.subreddit.flair.templates.update(
+                template['id'], 'updated by PRAW',
+                background_color='#000000')
 
 
 class TestSubredditLinkFlairTemplates(IntegrationTest):
@@ -409,7 +667,12 @@ class TestSubredditLinkFlairTemplates(IntegrationTest):
         with self.recorder.use_cassette(
                 'TestSubredditLinkFlairTemplates.test__iter'):
             templates = list(self.subreddit.flair.link_templates)
-        assert len(templates) == 1
+        assert len(templates) > 100
+
+        for template in templates:
+            assert template['id']
+            assert isinstance(template['richtext'], list)
+            assert all(isinstance(item, dict) for item in template['richtext'])
 
     @mock.patch('time.sleep', return_value=None)
     def test_add(self, _):
@@ -417,7 +680,17 @@ class TestSubredditLinkFlairTemplates(IntegrationTest):
         with self.recorder.use_cassette(
                 'TestSubredditLinkFlairTemplates.test_add'):
             for i in range(101):
-                self.subreddit.flair.link_templates.add('PRAW{}'.format(i))
+                self.subreddit.flair.link_templates.add('PRAW{}'.format(i),
+                                                        css_class='myCSS')
+
+    @mock.patch('time.sleep', return_value=None)
+    def test_add_v2(self, _):
+        self.reddit.read_only = False
+        with self.recorder.use_cassette(
+                'TestSubredditLinkFlairTemplates.test_add_v2'):
+            for i in range(101):
+                self.subreddit.flair.link_templates.add(
+                    'PRAW{}'.format(i), text_color='light')
 
     def test_clear(self):
         self.reddit.read_only = False
@@ -741,7 +1014,7 @@ class TestSubredditModmail(IntegrationTest):
     def test_conversations(self, _):
         self.reddit.read_only = False
         conversations = self.reddit.subreddit('all').modmail \
-                                                    .conversations()
+            .conversations()
         with self.recorder.use_cassette(
                 'TestSubredditModmail.test_conversations'):
             for conversation in conversations:
@@ -752,7 +1025,7 @@ class TestSubredditModmail(IntegrationTest):
     def test_conversations__params(self, _):
         self.reddit.read_only = False
         conversations = self.reddit.subreddit('all').modmail \
-                                                    .conversations(state='mod')
+            .conversations(state='mod')
         with self.recorder.use_cassette(
                 'TestSubredditModmail.test_conversations__params'):
             for conversation in conversations:
