@@ -3,6 +3,7 @@ import types
 import mock
 import pytest
 from praw import __version__, Reddit
+from praw.config import Config
 from praw.const import configparser
 from praw.exceptions import ClientException
 from prawcore import Requestor
@@ -77,6 +78,16 @@ class TestReddit(UnitTest):
             reddit.read_only = False
             assert not reddit.read_only
 
+    def test_read_only__with_authenticated_core__non_confidential(self):
+        with Reddit(client_id='dummy', client_secret=None,
+                    redirect_uri='dummy', user_agent='dummy',
+                    refresh_token='dummy') as reddit:
+            assert not reddit.read_only
+            reddit.read_only = True
+            assert reddit.read_only
+            reddit.read_only = False
+            assert not reddit.read_only
+
     def test_read_only__with_script_authenticated_core(self):
         with Reddit(password='dummy', username='dummy',
                     **self.REQUIRED_DUMMY_SETTINGS) as reddit:
@@ -112,7 +123,7 @@ class TestReddit(UnitTest):
         for setting in self.REQUIRED_DUMMY_SETTINGS:
             with pytest.raises(ClientException) as excinfo:
                 settings = self.REQUIRED_DUMMY_SETTINGS.copy()
-                del settings[setting]
+                settings[setting] = Config.CONFIG_NOT_SET
                 Reddit(**settings)
             assert str(excinfo.value).startswith('Required configuration '
                                                  'setting \'{}\' missing.'
